import binascii
from os import urandom

import humanize
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

from durin.settings import durin_settings
from durin.signals import token_renewed

User = settings.AUTH_USER_MODEL


def _create_token_string() -> str:
    return binascii.hexlify(
        urandom(int(durin_settings.TOKEN_CHARACTER_LENGTH / 2))
    ).decode()


class Client(models.Model):
    name = models.CharField(
        max_length=64,
        null=False,
        blank=False,
        db_index=True,
        unique=True,
        help_text=_("A unique identification name for the client."),
    )
    token_ttl = models.DurationField(
        null=False,
        default=durin_settings.DEFAULT_TOKEN_TTL,
        verbose_name=_("Token Time To Live (TTL)"),
        help_text=_(
            """
            Token Time To Live (TTL) in timedelta. Format: <em>DAYS HH:MM:SS</em>.
            """
        ),
    )

    def __str__(self):
        td = humanize.naturaldelta(self.token_ttl)
        return "({0}, {1})".format(self.name, td)


class AuthTokenManager(models.Manager):
    def create(self, user, client, delta_ttl=None):
        token = _create_token_string()

        if delta_ttl is not None:
            expiry = timezone.now() + delta_ttl
        else:
            expiry = timezone.now() + client.token_ttl

        instance = super(AuthTokenManager, self).create(
            token=token, user=user, client=client, expiry=expiry
        )
        return instance


class AuthToken(models.Model):
    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["user", "client"], name="unique token for user per client"
            )
        ]

    objects = AuthTokenManager()

    token = models.CharField(
        max_length=durin_settings.TOKEN_CHARACTER_LENGTH,
        null=False,
        blank=False,
        db_index=True,
        unique=True,
        help_text=_("Token is auto-generated on save."),
    )
    user = models.ForeignKey(
        User,
        null=False,
        blank=False,
        related_name="auth_token_set",
        on_delete=models.CASCADE,
    )
    client = models.ForeignKey(
        Client,
        null=False,
        blank=False,
        related_name="auth_token_set",
        on_delete=models.CASCADE,
    )
    created = models.DateTimeField(auto_now_add=True)
    expiry = models.DateTimeField(null=False)

    def renew_token(self, renewed_by):
        new_expiry = timezone.now() + self.client.token_ttl
        self.expiry = new_expiry
        self.save(update_fields=("expiry",))
        token_renewed.send(
            sender=renewed_by,
            username=self.user.get_username(),
            token_id=self.pk,
            expiry=new_expiry,
        )
        return new_expiry

    @property
    def expires_in(self) -> str:
        if self.expiry:
            td = self.expiry - self.created
            return humanize.naturaldelta(td)
        else:
            return "N/A"

    @property
    def has_expired(self) -> bool:
        return timezone.now() > self.expiry

    def __repr__(self) -> str:
        return "({0}, {1}/{2})".format(
            self.token, self.user.get_username(), self.client.name
        )

    def __str__(self) -> str:
        return self.token



from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header

from durin.models import AuthToken
from durin.settings import durin_settings
from durin.signals import token_expired

# try to import memoize
memoize = None

try:
    from memoize import memoize
except ImportError:
    pass


class TokenAuthentication(BaseAuthentication):
    """
    This authentication scheme uses Durin's
    :class:`durin.models.AuthToken` for authentication.

    Similar to `DRF's authentication system
    <http://www.django-rest-framework.org/api-guide/authentication/>`__,
    it overrides it a bit to
    accomodate that tokens can be expired.

    If successful,

    - ``request.user`` will be a django ``User`` instance
    - ``request.auth`` will be an ``AuthToken`` instance
    """

    model = AuthToken

    def authenticate(self, request):
        auth = get_authorization_header(request).split()
        prefix = durin_settings.AUTH_HEADER_PREFIX.encode()

        if not auth or auth[0].lower() != prefix.lower():
            return None
        if len(auth) == 1:
            msg = _("Invalid token header. No credentials provided.")
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _("Invalid token header. " "Token string should not contain spaces.")
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(auth[1])

    @classmethod
    def authenticate_credentials(cls, token):
        """
        Verify that the given token exists in the database
        """
        token = token.decode("utf-8")
        try:
            auth_token = AuthToken.objects.get(token=token)
            if cls._cleanup_token(auth_token):
                e = _("The given token has expired.")
                raise exceptions.AuthenticationFailed(e)
            return cls.validate_user(auth_token)
        except exceptions.AuthenticationFailed as e:
            raise exceptions.AuthenticationFailed(e)
        except Exception:
            msg = _("Invalid token.")
            raise exceptions.AuthenticationFailed(msg)

    @staticmethod
    def validate_user(auth_token: AuthToken):
        if not auth_token.user.is_active:
            raise exceptions.AuthenticationFailed(_("User inactive or deleted."))
        return (auth_token.user, auth_token)

    def authenticate_header(self, request):
        return durin_settings.AUTH_HEADER_PREFIX

    @classmethod
    def _cleanup_token(cls, auth_token: AuthToken):
        if auth_token.expiry is not None:
            if auth_token.has_expired:
                username = auth_token.user.get_username()
                auth_token.delete()
                token_expired.send(sender=cls, username=username, source="auth_token")
                return True
        return False


# if memoize is available, create another token authentication class
# which uses django-memoize for caching
if memoize:

    class CachedTokenAuthentication(TokenAuthentication):
        """
        Similar to ``TokenAuthentication`` but uses
        `django-memoize <https://pythonhosted.org/django-memoize/>`__
        as cache backend for faster lookups.

        The cache timeout is configurable by setting the
        ``REST_DURIN["TOKEN_CACHE_TIMEOUT"]`` under your app's ``settings.py``.

        **How To Enable:**

        1. Install django-memoize

        .. parsed-literal::
            pip install django-memoize

        2. Add ``'memoize'`` to ``INSTALLED_APPS``.
        3. Then you need to use ``CachedTokenAuthentication``
           instead of ``TokenAuthentication``.
        """

        @classmethod
        @memoize(timeout=int(durin_settings.TOKEN_CACHE_TIMEOUT))
        def authenticate_credentials(cls, token):
            return super().authenticate_credentials(token)

        def __repr__(self):
            return self.__class__.__name__



from django.conf.urls import url

from durin import views

urlpatterns = [
    url(r"login/", views.LoginView.as_view(), name="durin_login"),
    url(r"refresh/", views.RefreshView.as_view(), name="durin_refresh"),
    url(r"logout/", views.LogoutView.as_view(), name="durin_logout"),
    url(r"logoutall/", views.LogoutAllView.as_view(), name="durin_logoutall"),
]



from datetime import timedelta

from django.conf import settings
from django.test.signals import setting_changed
from rest_framework.settings import APISettings, api_settings

USER_SETTINGS = getattr(settings, "REST_DURIN", None)

DEFAULTS = {
    "DEFAULT_TOKEN_TTL": timedelta(days=1),
    "TOKEN_CHARACTER_LENGTH": 64,
    "USER_SERIALIZER": None,
    "AUTH_HEADER_PREFIX": "Token",
    "EXPIRY_DATETIME_FORMAT": api_settings.DATETIME_FORMAT,
    "TOKEN_CACHE_TIMEOUT": 60,
    "REFRESH_TOKEN_ON_LOGIN": False,
}

IMPORT_STRINGS = {
    "USER_SERIALIZER",
}

durin_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*args, **kwargs):
    global durin_settings
    setting, value = kwargs["setting"], kwargs["value"]
    if setting == "REST_DURIN":
        durin_settings = APISettings(value, DEFAULTS, IMPORT_STRINGS)


setting_changed.connect(reload_api_settings)



from django.contrib import admin

from durin import models


@admin.register(models.AuthToken)
class AuthTokenAdmin(admin.ModelAdmin):
    """Django's ModelAdmin for AuthToken.\n
    In most cases, you would want to override this to make
    ``AuthTokenAdmin.raw_id_fields = ("user",)``
    """

    exclude = ("token", "expiry")
    list_display = (
        "token",
        "client_name",
        "user",
        "created",
        "expires_in",
    )
    list_filter = ("client__name", "user")

    fieldsets = [
        (
            "API Auth Token",
            {
                "fields": ("user", "client"),
                "description": """
                    <h3>Token will be auto-generated on save.</h3>
                    <h3>Token will carry the same expiry as the
                     selected client's token TTL.</h3>
                """,
            },
        ),
    ]

    def client_name(self, obj):
        return obj.client.name

    def save_model(self, request, obj, form, change):
        return models.AuthToken.objects.create(obj.user, obj.client)


@admin.register(models.Client)
class ClientAdmin(admin.ModelAdmin):
    """
    Django's ModelAdmin for Client.
    """

    list_display = ("id", "name", "token_ttl")



import django.dispatch

token_expired = django.dispatch.Signal(providing_args=["username", "source"])
"""
When a token is expired and deleted.

        providing_args=["username", "source"]
"""

token_renewed = django.dispatch.Signal(
    providing_args=["username", "token_id", "expiry"]
)
"""
When a token is renewed by either :class:`durin.views.LoginView`
or :class:`durin.views.RefreshView`.

        providing_args=["username", "token_id", "expiry"]
"""



"""
# django-rest-durin

Per API client token authentication Module for django rest framework.

## Docs & Example Usage: https://github.com/eshaan7/django-rest-durin
"""
from setuptools import find_packages, setup

# Get the long description from the relevant file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

GITHUB_URL = "https://github.com/eshaan7/django-rest-durin"

setup(
    name="django-rest-durin",
    url=GITHUB_URL,
    version="0.1.0",
    license="MIT",
    description="""
    Per API client token authentication Module for django rest framework.
    """,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Eshaan Bansal",
    author_email="eshaan7bansal@gmail.com",
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP :: Session",
        "Environment :: Web Environment",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="django rest authentication login token client auth",
    packages=find_packages(exclude=[".github", "docs", "tests", "example_project"]),
    install_requires=["django>=2.2", "djangorestframework>=3.7.0", "humanize"],
    project_urls={
        "Documentation": "https://django-rest-durin.readthedocs.io/",
        "Funding": "https://www.paypal.me/eshaanbansal",
        "Source": GITHUB_URL,
        "Tracker": "{}/issues".format(GITHUB_URL),
    },
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": ["black==20.8b1", "flake8", "django-nose", "django-memoize", "isort"],
        "test": ["black==20.8b1", "flake8", "django-nose", "django-memoize", "isort"],
    },
)



from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.authtoken.serializers import AuthTokenSerializer
from rest_framework.exceptions import ParseError
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.serializers import DateTimeField
from rest_framework.views import APIView

from durin.auth import TokenAuthentication
from durin.models import AuthToken, Client
from durin.settings import durin_settings


class LoginView(APIView):
    """Durin's Login View.\n
    This view will return a JSON response when valid ``username``, ``password`` and
    if not overwritten ``client`` fields are POSTed to the view using
    form data or JSON.

    It uses the default serializer provided by
    Django-Rest-Framework (``rest_framework.authtoken.serializers.AuthTokenSerializer``)
    to validate the user credentials.

    It is possible to customize LoginView behaviour by overriding the following
    helper methods:
    """

    authentication_classes = []
    permission_classes = []

    def get_context(self):
        """
        to change the context passed to the ``UserSerializer``.
        """
        return {"request": self.request, "format": self.format_kwarg, "view": self}

    @staticmethod
    def validate_and_return_user(request):
        serializer = AuthTokenSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data["user"]

    @staticmethod
    def get_client_obj(request) -> "Client":
        """
        To get and return the associated :class:`durin.models.Client` object.
        """
        client_name = request.data.get("client", None)
        if not client_name:
            raise ParseError("No client specified.", status.HTTP_400_BAD_REQUEST)
        return Client.objects.get(name=client_name)

    @classmethod
    def get_token_obj(cls, request, client: "Client") -> "AuthToken":
        """
        Flow used to return the :class:`durin.models.AuthToken` object.
        """
        try:
            # a token for this user and client already exists, so we can just return it
            token = AuthToken.objects.get(user=request.user, client=client)
            if durin_settings.REFRESH_TOKEN_ON_LOGIN:
                cls.renew_token(token)
        except ObjectDoesNotExist:
            # create new token
            token = AuthToken.objects.create(request.user, client)

        return token

    @classmethod
    def renew_token(cls, token_obj: "AuthToken") -> None:
        """
        How to renew the token instance in case
        ``settings.REFRESH_TOKEN_ON_LOGIN`` is set to ``True``.
        """
        token_obj.renew_token(renewed_by=cls)

    @staticmethod
    def format_expiry_datetime(expiry):
        """
        To format the expiry ``datetime`` object at your convenience.
        """
        datetime_format = durin_settings.EXPIRY_DATETIME_FORMAT
        return DateTimeField(format=datetime_format).to_representation(expiry)

    def get_user_serializer_class(self):
        """
        To change the class used for serializing the user.
        """
        return durin_settings.USER_SERIALIZER

    def get_post_response_data(self, request, token_obj: "AuthToken"):
        """
        Override this to return a fully customized payload.
        """
        UserSerializer = self.get_user_serializer_class()
        data = {
            "expiry": self.format_expiry_datetime(token_obj.expiry),
            "token": token_obj.token,
        }
        if UserSerializer is not None:
            data["user"] = UserSerializer(request.user, context=self.get_context()).data
        return data

    def post(self, request, *args, **kwargs):
        request.user = self.validate_and_return_user(request)
        client = self.get_client_obj(request)
        token_obj = self.get_token_obj(request, client)
        user_logged_in.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        data = self.get_post_response_data(request, token_obj)
        return Response(data)


class RefreshView(APIView):
    """Durin's Refresh View\n
    This view accepts only a post request with an empty body.
    It responds to Durin Token Authentication. On a successful request,

    1. The given token's expiry is extended by it's associated
       :py:attr:`durin.models.Client.token_ttl`
       duration and a JSON object will be returned containing a single ``expiry``
       key as the new timestamp for when the token expires.

    2. :meth:`durin.signals.token_renewed` is called.
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    @staticmethod
    def format_expiry_datetime(expiry):
        """
        To format the expiry ``datetime`` object at your convenience.
        """
        datetime_format = durin_settings.EXPIRY_DATETIME_FORMAT
        return DateTimeField(format=datetime_format).to_representation(expiry)

    def post(self, request, *args, **kwargs):
        auth_token = request._auth
        new_expiry = auth_token.renew_token(renewed_by=self.__class__)
        new_expiry_repr = self.format_expiry_datetime(new_expiry)
        return Response({"expiry": new_expiry_repr}, status=status.HTTP_200_OK)


class LogoutView(APIView):
    """Durin's Logout View.\n
    This view accepts only a post request with an empty body.
    It responds to Durin Token Authentication. On a successful request,

    1. The token used to authenticate is deleted from
       the database and can no longer be used to authenticate.

    2. :meth:`django.contrib.auth.signals.user_logged_out` is called.

    :returns: 204 (No content)
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        request._auth.delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class LogoutAllView(APIView):
    """Durin's LogoutAllView.\n
    This view accepts only a post request with an empty body. It responds to Durin Token
    Authentication.
    On a successful request,

    1. The token used to authenticate, and **all other tokens**
       registered to the same ``User`` account, are deleted from the
       system and can no longer be used to authenticate.

    2. :meth:`django.contrib.auth.signals.user_logged_out` is called.

    :returns: 204 (No content)
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        request.user.auth_token_set.all().delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return Response(None, status=status.HTTP_204_NO_CONTENT)
